local _, BPGP = ...

local assert, next, pairs, select, type, unpack = assert, next, pairs, select, type, unpack
local tinsert, wipe = tinsert, wipe

local locked = false

local eventCallbacks = {}
local callbacksCache = {}

local eventQueue = {}
local pushPointer = 0
local shiftPointer = 0

local unitEventFrames = {}
local registeredUnitEvents = {}

-- High-performance event queue for simultaneous events processing in FIFO order

-- Adds event to the end of the queue
local function PushEvent(...)
  pushPointer = pushPointer + 1
  local eventStackHandle = eventQueue[pushPointer]
  if not eventStackHandle then
    eventStackHandle = {}
    eventQueue[pushPointer] = eventStackHandle
  end
	for i = 1, select("#", ...) do
    eventStackHandle[i] = select(i, ...)
	end
end

-- Wipes unpacked table for recycle
local function WipeHandle(handle, ...)
  wipe(handle)
  return ...
end

-- Retrieves the event from the beginning of the queue, resets pointers on reaching the end
local function ShiftEvent()
  shiftPointer = shiftPointer + 1
  local eventStackHandle = eventQueue[shiftPointer]
  if shiftPointer == pushPointer then
    pushPointer = 0
    shiftPointer = 0
  end
  return WipeHandle(eventStackHandle, unpack(eventStackHandle))
end

-- Event handling

local function HandleEvent(event, ...)
  wipe(callbacksCache)
	local callbacks = eventCallbacks[event]
	if callbacks then
    for callback in pairs(callbacks) do
      tinsert(callbacksCache, callback)
    end
		for i = 1, #callbacksCache do
			callbacksCache[i](event, ...)
		end
	end
end

local function OnEvent(_, event, ...)
  if locked then
    PushEvent(event, ...)
    return
  end
  locked = true
  HandleEvent(event, ...)
	while pushPointer > 0 do
		HandleEvent(ShiftEvent())
	end
  locked = false
end

local EventFrame = CreateFrame("Frame")
EventFrame:SetScript("OnEvent", OnEvent)

-- Core methods

local function RegisterUnitEvent(event, ...)
  registeredUnitEvents[event] = true
  for i = 1, select("#", ...) do
    local unit = select(i, ...)
    if not unit then break end
    assert(type(unit) == "string", "Invalid UnitID!")
    local frame = unitEventFrames[unit]
    if not frame then
      frame = CreateFrame("Frame")
      frame:SetScript("OnEvent", OnEvent)
      unitEventFrames[unit] = frame
    end
    frame:RegisterUnitEvent(event, unit)
  end
end

local function UnregisterUnitEvent(event, ...)
  registeredUnitEvents[event] = nil
  for i = 1, select("#", ...) do
    local unit = select(i, ...)
    if not unit then break end
    assert(type(unit) == "string", "Invalid UnitID!")
    local frame = unitEventFrames[unit]
    if frame then
      frame:UnregisterEvent(event)
    end
  end
end

local function RegisterEvent(event, callback, ...)
  assert(type(event) == "string" and type(callback) == "function")
	if not eventCallbacks[event] then
		eventCallbacks[event] = {}
	end
  if event:sub(0, 7) == "CUSTOM_" then
    -- No actual registration needed, custom event is just queue-delayed callback
  elseif event:sub(0, 5) == "UNIT_" then
    if select("#", ...) > 0 then
      assert(not next(eventCallbacks[event]), "BPGP Event Handler supports only 1 registration per filtered unit event!")
      -- Special registration for filtered unit event, we'll use separate frame for each unit specified
      RegisterUnitEvent(event, ...)
    else
      assert(not registeredUnitEvents[event], "BPGP Event Handler doesn't support mixed registrations for same unit event!")
      EventFrame:RegisterEvent(event)
    end
  else
    EventFrame:RegisterEvent(event)
  end
	eventCallbacks[event][callback] = true
end

local function UnregisterEvent(event, callback, ...)
  assert(type(event) == "string" and type(callback) == "function")
	local callbacks = eventCallbacks[event]
	if callbacks and callbacks[callback] then
		callbacks[callback] = nil
		if not next(callbacks) then
			eventCallbacks[event] = nil
      if event:sub(0, 7) == "CUSTOM_" then
        -- No actual unregistration needed, custom event is just queue-delayed callback
      elseif event:sub(0, 5) == "UNIT_" and select("#", ...) > 0 then
        -- Special unregistration for filtered unit event
        UnregisterUnitEvent(event, ...)
      else
        EventFrame:UnregisterEvent(event)
      end
		end
	end
end

local function FireEvent(event, ...)
  OnEvent(nil, event, ...)
end

local function RegisterCallback(event, callback, ...)
  RegisterEvent("CUSTOM_"..event, callback, ...)
end

local function UnregisterCallback(event, callback, ...)
  UnregisterEvent("CUSTOM_"..event, callback, ...)
end

local function Fire(event, ...)
  FireEvent("CUSTOM_"..event, ...)
end

local function EmulateEvent(event, ...)
  for _, frame in pairs({GetFramesRegisteredForEvent(event)}) do
    local func = frame:GetScript('OnEvent')
    pcall(func, frame, event, ...)
  end
end

----------------------------------------
-- Public interface
----------------------------------------

local function Embed(target)
  target.RegisterEvent = RegisterEvent
  target.UnregisterEvent = UnregisterEvent
  target.FireEvent = FireEvent
  target.EmulateEvent = EmulateEvent
  target.RegisterCallback = RegisterCallback
  target.UnregisterCallback = UnregisterCallback
  target.Fire = Fire
end

Embed(BPGP)
