local _, BPGP = ...

local DistributionTracker = BPGP.NewModule("DistributionTracker")

local L = LibStub("AceLocale-3.0"):GetLocale("BPGP")
local DLG = LibStub("LibDialog-1.0")
local LibDeformat = LibStub("LibDeformat-3.0")
LibStub("AceComm-3.0"):Embed(DistributionTracker)

local Debugger = BPGP.Debugger
local Common = BPGP.GetLibrary("Common")
local Coroutine = BPGP.GetLibrary("Coroutine")

local DataManager = BPGP.GetModule("DataManager")
local MasterLooter = BPGP.GetModule("MasterLooter")

local private = {
  dbDefaults = {
    profile = {
      enabled = false, -- Loading is managed by MasterLooter module
      cache = {
        mode = nil,
        itemLink = nil,
        responses = {},
      },
    },
  },
  decoder = {L["Need"], L["Roll"], L["Greed"], L["Pass"], L["Auto Pass"], L["Waiting..."]},
  equippableMultiSlots = {
    ["INVTYPE_FINGER"] = true,
    ["INVTYPE_TRINKET"] = true,
    ["INVTYPE_WEAPON"] = true,
    ["INVTYPE_BAG"] = true,
  },
  priorities = {1, 1, 2, 4, 5, 3},
  distributionStartFormat = "%d\31%s\31%s",
}
Debugger.Attach("DistributionTracker", private)

function DistributionTracker:OnAddonLoad()
  self.db = BPGP.db:RegisterNamespace("DistributionTracker", private.dbDefaults)
end

function DistributionTracker:OnEnable()
  Debugger("DistributionTracker:OnEnable()")
  self:RegisterComm("BPGPDISTRSTART", private.HandleDistributionStartComm)
  self:RegisterComm("BPGPDISTRRESP", private.HandleDistributionResponseComm)
  self:RegisterComm("BPGPDISTRSTOP", private.HandleDistributionStopComm)
  if not DataManager.SelfInRaid() then
    self:Disable()
  end
end

function DistributionTracker:OnDisable()
  Debugger("DistributionTracker:OnDisable()")
  self:UnregisterComm("BPGPDISTRSTART", private.HandleDistributionStartComm)
  self:UnregisterComm("BPGPDISTRRESP", private.HandleDistributionResponseComm)
  self:UnregisterComm("BPGPDISTRSTOP", private.HandleDistributionStopComm)
  self.ResetDistributionResults()
end

----------------------------------------
-- Event handlers
----------------------------------------

function private.HandleDistributionStartComm(prefix, message, distribution, sender)
  if not DataManager.IsInRaid(sender) then return end
  local mode, itemLink, comment = LibDeformat(message, private.distributionStartFormat)
  local itemId = Common:GetItemId(itemLink)
  if not mode or not itemId then return end
  DistributionTracker.ResetDistributionResults()
  DistributionTracker.db.profile.cache.mode = mode
  DistributionTracker.db.profile.cache.itemLink = itemLink
  if mode == 1 then
    BPGP.GetModule("RollTracker").StartTracking(itemLink)
  elseif mode == 2 or mode == 3 then
    BPGP.GetModule("RollTracker").StartTracking(itemLink)
    Coroutine:RunAsync(private.AsyncHandleDistributionStartComm, mode, itemId, itemLink, comment)
  end
  BPGP.GetModule("StandingsManager").DestroyStandings()
end

function private.AsyncHandleDistributionStartComm(mode, itemId, incomingItemLink, comment)
  local itemLink = nil
  for i = 1, 6 do
    itemLink = select(2, GetItemInfo(incomingItemLink))
    local parserState = Common:LoadParser(itemLink)
    if i == 6 then break end
    if parserState >= 2 then break end
    Coroutine:Sleep(0.05)
    if DataManager.SelfML() then break end
  end
  local autopass, reason = private.ShouldAutopass(itemLink)
  Common:UnloadParser(itemLink)
  if autopass then
    BPGP.Print(L["Auto Pass: %s (%s)"]:format(itemLink, reason))
    DistributionTracker.ResponseDistribution(incomingItemLink, 5)
  else
    Coroutine:RunAsync(private.AsyncShowDistributionPopup, mode, itemId, incomingItemLink, comment)
  end
end

function private.ShouldAutopass(itemLink)
  if BPGP.GetModule("GUI").db.profile.autoPassMode == "HARD_PASS" then
    return true, L["hard pass mode enabled"]
  end
  if Common:IsRecipe(itemLink) then
    return false
  end
  local playerClass, classLimit = DataManager.GetSelfClass(), Common:ParseClasses(itemLink)
  if #classLimit > 0 and not Common:FindNext(classLimit, playerClass) then
    return true, L["it's not intended for %s"]:format(DataManager.GetSelfClassLocalized():lower())
  end
  if not IsEquippableItem(itemLink) then
    return false
  end
  if not Common:CanEquipItem(itemLink, playerClass) then
    return true, L["it's not equippable by %s"]:format(DataManager.GetSelfClassLocalized():lower())
  end
  if Common:IsNotDesignedForClass(itemLink, playerClass) then
    return true, L["it's useless for %s"]:format(DataManager.GetSelfClassLocalized():lower())
  end
  local itemCount = GetItemCount(itemLink, true)
  if itemCount > 0 then
    local isUnique = Common:ParseUnique(itemLink)
    if isUnique then
      return true, L["you already have that unique item"]
    else
      local slotName = select(9, GetItemInfo(itemLink))
      if private.equippableMultiSlots[slotName] then
        if itemCount >= 2 and slotName ~= "INVTYPE_BAG" then
          return true, L["you already have %d of that items"]:format(itemCount)
        end
      else
        return true, L["you already have that item"]
      end
    end
  end
  return false
end

function private.AsyncShowDistributionPopup(mode, itemId, itemLink, comment)
  while DLG:ActiveDialog("BPGP_ITEM_DISTRIBUTION") do
    Coroutine:Sleep(0.1)
  end
  DLG:Spawn("BPGP_ITEM_DISTRIBUTION", {mode = mode, item = itemLink, icon = GetItemIcon(itemId), comment = comment})
end

function private.HandleDistributionResponseComm(prefix, message, distribution, sender)
  if not DataManager.IsInRaid(sender) then return end
  local itemLink, response, itemLinks = strsplit(";", message)
  if itemLink ~= DistributionTracker.db.profile.cache.itemLink then return end
  local equipped1, equipped2 = strsplit(",", itemLinks or "")
  DistributionTracker.RegisterDistributionResponse(itemLink, sender, response, equipped1, equipped2)
end

function private.HandleDistributionStopComm(prefix, message, distribution, sender)
  if not DataManager.IsInRaid(sender) or DataManager.SelfML() then return end
  DistributionTracker.ResetDistributionResults()
  BPGP.GetModule("RollTracker").StopTracking(message)
end

function DistributionTracker.HandleSimpleRoll(playerName)
  local response, equipped1, equipped2 = DistributionTracker.db.profile.cache.responses[playerName], nil, nil
  if response then 
    response, equipped1, equipped2 = unpack(response)
    if response < 4 then 
      return
    end
  end
  DistributionTracker.RegisterDistributionResponse(DistributionTracker.db.profile.cache.itemLink, playerName, 2, equipped1, equipped2)
end

----------------------------------------
-- Public interface
----------------------------------------

function DistributionTracker.ResetDistributionResults()
  if not DistributionTracker.db.profile.cache.itemLink then return end
  DistributionTracker.db.profile.cache.mode = nil
  DistributionTracker.db.profile.cache.itemLink = nil
  table.wipe(DistributionTracker.db.profile.cache.responses)
  BPGP.GetModule("RollTracker").ResetRollResults()
  BPGP.GetModule("StandingsManager").DestroyStandings()
end

function DistributionTracker.StartDistribution(mode, itemLink, comment)
  if not DataManager.SelfML() then return end
  if not comment or comment == "" then comment = "<NONE>" end
  local message = string.format(private.distributionStartFormat, mode, itemLink, comment)
  DistributionTracker:SendCommMessage("BPGPDISTRSTART", message, "RAID", nil, "ALERT")
end

function DistributionTracker.ResponseDistribution(itemLink, response)
  if not DataManager.SelfInRaid() then return end
  if itemLink ~= DistributionTracker.db.profile.cache.itemLink then
    BPGP.Print(L["Distribution of %s has already ended!"]:format(itemLink))
    return
  end  
  local message = string.format("%s;%d;", itemLink, response)
  if response == 1 or response == 3 then -- Need or Greed
    RandomRoll(1, 100)
  end
  if IsEquippableItem(itemLink) then
    local slotName = select(9, GetItemInfo(itemLink))
    local itemLinks = Common:GetEquippedItemLinks(slotName)
    if itemLinks then message = message..table.concat(itemLinks, ",") end
  end
  DistributionTracker:SendCommMessage("BPGPDISTRRESP", message, "RAID", nil, "BULK")
end

function DistributionTracker.RegisterDistributionResponse(itemLink, playerName, response, equipped1, equipped2)
  DistributionTracker.db.profile.cache.responses[playerName] = {tonumber(response), equipped1, equipped2}
  BPGP.GetModule("StandingsManager").DestroyStandings()
end

function DistributionTracker.GetDistributionResponse(playerName)
  if not DistributionTracker.db.profile.cache.itemLink then return end
  local result = DistributionTracker.db.profile.cache.responses[playerName]
  if result then
    return unpack(result)
  else
    return 6
  end
end

function DistributionTracker.GetDistributionPriority(playerName)
  return private.priorities[DistributionTracker.GetDistributionResponse(playerName)]
end

function DistributionTracker.GetDecodedDistributionResponse(playerName)
  local response, equipped1, equipped2 = DistributionTracker.GetDistributionResponse(playerName)
  return private.decoder[response], equipped1, equipped2
end

function DistributionTracker.StopDistribution(itemLink)
  if itemLink and itemLink ~= DistributionTracker.db.profile.cache.itemLink then return end
  if not itemLink then itemLink = DistributionTracker.db.profile.cache.itemLink end
  if itemLink and DataManager.SelfML() then
    DistributionTracker:SendCommMessage("BPGPDISTRSTOP", itemLink, "RAID", nil, "ALERT")
  end
  DistributionTracker.ResetDistributionResults()
  BPGP.GetModule("RollTracker").StopTracking(itemLink)
end
